@@ -2,17 +2,31 @@
 // for details. All rights reserved. Use of this source code is governed by a
 // BSD-style license that can be found in the LICENSE file.
 
-// @dart = 2.6
-
 part of _interceptors;
 
-/// The super interceptor class for [JSInt] and [JSDouble]. The compiler
-/// recognizes this class as an interceptor, and changes references to
-/// [:this:] to actually use the receiver of the method, which is
-/// generated as an extra argument added to each member.
+/// Interceptor class for all Dart [num] implementations.
 ///
-/// Note that none of the methods here delegate to a method defined on JSInt or
-/// JSDouble.  This is exploited in [tryComputeConstantInterceptor].
+/// JavaScript numbers (doubles) are used to represent both Dart [int] and Dart
+/// [double] values. Some values, e.g. `3.0` are both Dart [int] values and Dart
+/// [double] values. Other values are just [double] values, e.g. `3.1`.
+///
+/// There are two disjoint subclasses of [JSNumber]: [JSInt] and [JSNumNotInt].
+///
+/// Most methods are on [JSNumber]. Since some values can 'be' (i.e. implement)
+/// both [int] and [double], the int and double operations have to be the same.
+/// Consider the JavaScript value `0`. This is both Dart int 0, and Dart double
+/// 0.0. From the dynamic type we can't tell the intention, so the
+/// `0.0.toString()` on the web returns `0`, and not `0.0` like on the Dart VM
+/// implementation. For `toString` we prefer the `int` version. For negation, we
+/// prefer the `double` version (returning `-0.0`, not `0`). This is usually
+/// hidden because the JavaScript `-0.0` value is also considered to implement
+/// [int].
+///
+/// Note that none of the methods in [JSNumber] delegate to a method defined on
+/// JSInt (or JSNumNotInt).  This is exploited in
+/// [tryComputeConstantInterceptor] to avoid most interceptor lookups on
+/// numbers.
+
 class JSNumber extends Interceptor implements double {
   const JSNumber();
 
@@ -68,7 +82,11 @@ class JSNumber extends Interceptor implements double {
       r'Math.abs(#)',
       this);
 
-  JSNumber get sign => this > 0 ? 1 : this < 0 ? -1 : this;
+  JSNumber get sign => (this > 0
+      ? 1
+      : this < 0
+          ? -1
+          : this) as JSNumber;
 
   static const int _MIN_INT32 = -0x80000000;
   static const int _MAX_INT32 = 0x7FFFFFFF;
@@ -183,7 +201,7 @@ class JSNumber extends Interceptor implements double {
     return result;
   }
 
-  String toStringAsExponential([int fractionDigits]) {
+  String toStringAsExponential([int? fractionDigits]) {
     String result;
     if (fractionDigits != null) {
       checkInt(fractionDigits);
@@ -224,7 +242,7 @@ class JSNumber extends Interceptor implements double {
   static String _handleIEtoString(String result) {
     // Result is probably IE's untraditional format for large numbers,
     // e.g., "8.0000000000008(e+15)" for 0x8000000000000800.toString(16).
-    var match = JS('JSArray|Null',
+    List? match = JS('JSArray|Null',
         r'/^([\da-z]+)(?:\.([\da-z]+))?\(e\+(\d+)\)$/.exec(#)', result);
     if (match == null) {
       // Then we don't know how to handle it at all.
@@ -234,7 +252,7 @@ class JSNumber extends Interceptor implements double {
     int exponent = JS('int', '+#', match[3]);
     if (match[2] != null) {
       result = JS('String', '# + #', result, match[2]);
-      exponent -= JS('int', '#.length', match[2]);
+      exponent -= JS<int>('int', '#.length', match[2]);
     }
     return result + '0' * exponent;
   }
@@ -308,13 +326,13 @@ class JSNumber extends Interceptor implements double {
   JSNumber operator %(num other) {
     if (other is! num) throw argumentErrorValue(other);
     // Euclidean Modulo.
-    num result = JS('num', r'# % #', this, other);
+    JSNumber result = JS<JSNumber>('JSNumber', r'# % #', this, other);
     if (result == 0) return JS('num', '0'); // Make sure we don't return -0.0.
     if (result > 0) return result;
-    if (JS('num', '#', other) < 0) {
-      return result - JS('num', '#', other);
+    if (other < 0) {
+      return JS<JSNumber>('JSNumber', '# - #', result, other);
     } else {
-      return result + JS('num', '#', other);
+      return JS<JSNumber>('JSNumber', '# + #', result, other);
     }
   }
 
@@ -339,7 +357,7 @@ class JSNumber extends Interceptor implements double {
   }
 
   int _tdivSlow(num other) {
-    var quotient = JS('num', r'# / #', this, other);
+    num quotient = JS('num', r'# / #', this, other);
     if (quotient >= _MIN_INT32 && quotient <= _MAX_INT32) {
       // This path includes -0.0 and +0.0.
       return JS('int', '# | 0', quotient);
@@ -367,7 +385,7 @@ class JSNumber extends Interceptor implements double {
 
   num operator <<(num other) {
     if (other is! num) throw argumentErrorValue(other);
-    if (JS('num', '#', other) < 0) throw argumentErrorValue(other);
+    if (other < 0) throw argumentErrorValue(other);
     return _shlPositive(other);
   }
 
@@ -380,14 +398,14 @@ class JSNumber extends Interceptor implements double {
   }
 
   num operator >>(num other) {
-    if (false) _shrReceiverPositive(other);
     if (other is! num) throw argumentErrorValue(other);
-    if (JS('num', '#', other) < 0) throw argumentErrorValue(other);
+    if (other < 0) throw argumentErrorValue(other);
+    if (false) _shrReceiverPositive(other);
     return _shrOtherPositive(other);
   }
 
   num _shrOtherPositive(num other) {
-    return JS('num', '#', this) > 0
+    return this > 0
         ? _shrBothPositive(other)
         // For negative numbers we just clamp the shift-by amount.
         // `this` could be negative but not have its 31st bit set.
@@ -397,7 +415,7 @@ class JSNumber extends Interceptor implements double {
   }
 
   num _shrReceiverPositive(num other) {
-    if (JS('num', '#', other) < 0) throw argumentErrorValue(other);
+    if (0 > other) throw argumentErrorValue(other);
     return _shrBothPositive(other);
   }
 
@@ -413,6 +431,17 @@ class JSNumber extends Interceptor implements double {
         : JS('JSUInt32', r'# >>> #', this, other);
   }
 
+  num operator >>>(num other) {
+    if (other is! num) throw argumentErrorValue(other);
+    if (other < 0) throw argumentErrorValue(other);
+    return _shruOtherPositive(other);
+  }
+
+  num _shruOtherPositive(num other) {
+    if (other > 31) return 0;
+    return JS('JSUInt32', r'# >>> #', this, other);
+  }
+
   num operator &(num other) {
     if (other is! num) throw argumentErrorValue(other);
     return JS('JSUInt32', r'(# & #) >>> 0', this, other);
@@ -469,7 +498,11 @@ class JSInt extends JSNumber implements int {
       this);
 
   @override
-  JSInt get sign => this > 0 ? 1 : this < 0 ? -1 : this;
+  JSInt get sign => (this > 0
+      ? 1
+      : this < 0
+          ? -1
+          : this) as JSInt;
 
   @override
   JSInt operator -() => JS('int', r'-#', this);
@@ -488,7 +521,7 @@ class JSInt extends JSNumber implements int {
   }
 
   int get bitLength {
-    JSInt nonneg = this < 0 ? -this - 1 : this;
+    int nonneg = JS<int>('int', '#', this < 0 ? -this - 1 : this);
     int wordBits = 32;
     while (nonneg >= 0x100000000) {
       nonneg = nonneg ~/ 0x100000000;
@@ -498,8 +531,7 @@ class JSInt extends JSNumber implements int {
   }
 
   static int _clz32(int uint32) {
-    // TODO(sra): Use `Math.clz32(uint32)` (not available on IE11).
-    return 32 - _bitCount(_spread(uint32));
+    return JS('JSUInt31', 'Math.clz32(#)', uint32);
   }
 
   // Returns pow(this, e) % m.
@@ -651,54 +683,14 @@ class JSInt extends JSNumber implements int {
     return _binaryGcd(x, y, false);
   }
 
-  // Assumes i is <= 32-bit and unsigned.
-  static int _bitCount(int i) {
-    // See "Hacker's Delight", section 5-1, "Counting 1-Bits".
-
-    // The basic strategy is to use "divide and conquer" to
-    // add pairs (then quads, etc.) of bits together to obtain
-    // sub-counts.
-    //
-    // A straightforward approach would look like:
-    //
-    // i = (i & 0x55555555) + ((i >>  1) & 0x55555555);
-    // i = (i & 0x33333333) + ((i >>  2) & 0x33333333);
-    // i = (i & 0x0F0F0F0F) + ((i >>  4) & 0x0F0F0F0F);
-    // i = (i & 0x00FF00FF) + ((i >>  8) & 0x00FF00FF);
-    // i = (i & 0x0000FFFF) + ((i >> 16) & 0x0000FFFF);
-    //
-    // The code below removes unnecessary &'s and uses a
-    // trick to remove one instruction in the first line.
-
-    i = _shru(i, 0) - (_shru(i, 1) & 0x55555555);
-    i = (i & 0x33333333) + (_shru(i, 2) & 0x33333333);
-    i = 0x0F0F0F0F & (i + _shru(i, 4));
-    i += _shru(i, 8);
-    i += _shru(i, 16);
-    return (i & 0x0000003F);
-  }
-
-  static int _shru(int value, int shift) => JS('int', '# >>> #', value, shift);
-  static int _shrs(int value, int shift) => JS('int', '# >> #', value, shift);
-  static int _ors(int a, int b) => JS('int', '# | #', a, b);
-
-  // Assumes i is <= 32-bit
-  static int _spread(int i) {
-    i = _ors(i, _shrs(i, 1));
-    i = _ors(i, _shrs(i, 2));
-    i = _ors(i, _shrs(i, 4));
-    i = _ors(i, _shrs(i, 8));
-    i = _shru(_ors(i, _shrs(i, 16)), 0);
-    return i;
-  }
-
   Type get runtimeType => int;
 
   int operator ~() => JS('JSUInt32', r'(~#) >>> 0', this);
 }
 
-class JSDouble extends JSNumber implements double {
-  const JSDouble();
+/// Interceptor for JavaScript values that are not a subclass of [JSInt].
+class JSNumNotInt extends JSNumber implements double {
+  const JSNumNotInt();
   Type get runtimeType => double;
 }
 
